;;;-*- Mode: Lisp; Package: SQL -*-

;; Portable SQL module
;; Version 0.9
;; Copyright (C) Paul Meurer 1999, 2000. All rights reserved.
;; paul.meurer@hit.uib.no

;;; "This software is based on the design of the ODBC interface provided by 
;;;  Harlequin Group plc in the LispWorks and LispWorks for the Windows 
;;;  Operating System products. This implementation is not the property of
;;;  Harlequin and they have no responsibility for its content or accuracy."

;; Documentation and the license agreement can be found in file 
;; "sql-odbc-documentation.lisp".
;; Bug reports and suggestions are highly welcome.


(in-package :sql)

(defparameter *default-database-type* :odbc)
(defparameter *trace-sql* nil
  "If set to a true value, the sql commands sent to the database are additionally
written to *TRACE-OUTPUT*, or, in the case *TRACE-SQL* is bound to an output stream,
to that stream. Useful for debugging.")

(defun initialize-database-type (&optional (database-type *default-database-type*)
                                               (set-default t))
  (when set-default (setf *default-database-type* database-type))
  (ecase database-type
    (:odbc
     #+:mk (mk:load-system :odbc)  
     #+(and (not :mk) :lispworks) (lw:load-system :odbc)               ;DIV 2001-Nov-26
     (use-package :odbc))
    (:oci
     #+:mk (mk:load-system :oci)))
  database-type)

(defmacro with-database ((database) &body body)
  "Binds *DEFAULT-DATABASE* to DATABASE in the scope of BODY"
  `(let ((*default-database* ,database))
     ,@body))

;; one of :new :warn-new :error :warn-old :old
(defparameter *connect-if-exists* :old)

;; should this be something different from a list?
(defvar *connected-databases* () "A list of connected databases.")

#+mcl
(defmethod initialize-instance ((query query) &key &allow-other-keys)
  (call-next-method)
  (terminate-when-unreachable query))

(defun disconnect-all-databases ()
  (dolist (database *connected-databases*)
    (disconnect :database database)))

#+mcl
(pushnew #'disconnect-all-databases *lisp-cleanup-functions*)

#+lispworks                   ;; DIV 2002-Jun-15
(lw:define-action "When quitting image" "Disconnect all databases"
                  'disconnect-all-databases
                  :once) ; because "When quitting image" is actually run twice

;; Support for database types is added by defining methods for the following
;; generic functions (there is no need to change the sql module, besides 
;; changing initialize-database-type):

(defgeneric db-connect (db-or-type db-name user-id password autocommit))
(defgeneric db-disconnect (database))
(defgeneric db-begin-transaction (database)) ;; not needed for odbc?
(defgeneric db-commit (database))
(defgeneric db-rollback (database))
(defgeneric db-query (database query-expression &optional flatp))
(defgeneric db-open-query (database query-expression &key &allow-other-keys))
(defgeneric db-close-query (database &key &allow-other-keys))
(defgeneric db-cancel-query (database))
(defgeneric db-fetch-query-results (database &optional count flatp))
(defgeneric db-execute-command (database sql-expression))
;; those two could well be one function
(defgeneric db-execute-parameterized (query sql-expression &rest parameters))
;(defgeneric db-query-parameterized (query sql-expression &rest parameters))
(defgeneric %db-execute (database sql-expression &key &allow-other-keys))
(defgeneric %read-query-data (database ignore-columns))
(defgeneric %initialize-query (database &optional arglen col-positions))
(defgeneric %db-bind-execute (query &rest parameters))
(defgeneric %db-reset-query (query))
(defgeneric db-map-query (database type function query-expression))
(defgeneric db-map-bind-query (query type function &rest parameters))
(defgeneric db-prepare-statement (database sql-expression &key &allow-other-keys))
(defgeneric db-describe-columns (database table-qualifier table-owner 
                                              table-name column-name))
(defgeneric db-data-sources (db-type))
(defgeneric column-names (query)) ;; returns a vector

;(defgeneric close-query (query))

;; dummy method? don't remember
(defmethod db-connect ((db-type (eql :rest)) db-name user-id password autocommit)
  (declare (ignore db-name password autocommit))
  (values user-id db-type))

;; dummy for odbc
(defmethod db-begin-transaction ((database database))
  nil)

(defmacro with-connection ((database &key name user-id password
                                        (db-type '*default-database-type*)
                                        (bind-as-default t)
                                        autocommit)
                              &body body)
  "Establishes a database connection. The connection is open in the scope of the body
and closed on exit of the macro. Arguments:
 DATABASE : symbol the new database connection object is bound to in the body of the macro.
 NAME, USER-ID, PASSWORD, DB-TYPE, AUTOCOMMIT: same keyword arguments as in CONNECT.
 BIND-AS-DEFAULT: If true, *DEFAULT-DATABASE* is bound to the database object in the scope
of the body."
  (let ((successp (gensym)))
    `(let* ((,database (connect ,name
                                :user-id ,user-id
                                :password ,password
                                :if-exists :new
                                :db-type ,db-type
                                :make-default nil
                                :autocommit ,autocommit))
            (*default-database* (if ,bind-as-default ,database *default-database*))
            (,successp nil))
       (declare (ignorable ,database ,successp))
       (unwind-protect
         (multiple-value-prog1 
           (progn ,@body)
           (setf ,successp t))
         (disconnect :database ,database)))))

(defun connect (db-name &key (user-id "") (password "")
                          (if-exists *connect-if-exists*) 
                          (db-type *default-database-type*)
                          (make-default t)
                          autocommit
                          &allow-other-keys)
  "Connects to a SQL database. Arguments:
 DB-NAME : Name of database to be used in connection string.
 USER-ID : Username
 PASSWORD : Password
 IF-EXISTS : Default: *connect-if-exists*.
  One of 
  :NEW : A new connection is created regardless of prior connections existing. 
  :WARN-NEW : Like :new, but a warning is issued. 
  :OLD : If an old connection exists, it is used, otherwise a new one is created. 
  :WARN-OLD : Like :OLD, but a warning is issued if an old connection is used. 
  :ERROR : If a connection already exists, an error is signalled.
 DB-TYPE : A database type (e.g., :odbc). Default: *default-database-type*.
 AUTOCOMMIT : Whether to open connection in autocommit mode. Default: nil.
 MAKE-DEFAULT : If non-nil, *default-database* is set to the database object.
Returns the database object or NIL if can't be connected."   ; DIV 2002-Jul-10
  (let ((db ;; check if connection of requested type already exists
         (dolist (db *connected-databases*)
           (when (and (eq db-type (db-type db))
                      (string= db-name (db-name db))
                      (string= user-id (db-user-id db)))
             (return db)))))
    (cond ((or (not db) 
               (eq if-exists :new)
               (eq if-exists :warn-new))
           (when (and db (eq if-exists :warn-new))
             (warn "A connection to '~a/~a' already exists." db-name user-id))
           (setf db (db-connect db-type db-name user-id password autocommit))
           (when (db-connected-p db)      ; DIV 2002-Jul-10 Push to list only if really so!
             (when (or make-default (null *default-database*))
               (setf *default-database* db))
             (push db *connected-databases*)
             db))
          ((eq if-exists :error)
           (error "A connection to '~a/~a' already exists." db-name user-id))
          ((or (eq if-exists :old)
               (eq if-exists :warn-old)) 
           (when (eq if-exists :warn-old)
             (warn "Using existing connection to '~a/~a'." db-name user-id))
           (unless (db-connected-p db)
             (db-connect db db-name user-id password autocommit))
           (when (db-connected-p db)                           ; DIV 2002-Jul-10
             (when (or make-default (null *default-database*))
               (setf *default-database* db))
             db)))))

(defun disconnect (&key (database *default-database*) (rollback t))
  "Disconnects from a database.
If rollback is t (the default), rolls back uncommited transactions."
  (when (null database)
    (error "No database is connected or no default database specified."))
  (when rollback (db-rollback database))
  (db-disconnect database))

(defun commit (&key (database *default-database*))
  "Commits pending transactions."
  (db-commit database))

(defun rollback (&key (database *default-database*))
  "Rolls back pending transactions."
  (db-rollback database))

;;; LMH ensure connection
(defun ensure-connection (database)
  (unless database (error "No connection to a database; call sql:connect first.")))

;;; LMH ensure connection
(defmethod execute-query :before (query-string &key flatp)
  (declare (ignore query-string))
  (ensure-connection *default-database*))

(defmethod execute-query ((query-string string) &key flatp)
  "Executes a SQL query string and fetches all rows."
  (db-query *default-database* query-string flatp))

(defmethod execute-query ((query-exp sql-query-expression) &key flatp)
  "Executes a SQL query expression and fetches all rows."
  (db-query *default-database* (sql-string query-exp) flatp))

;;; LMH ensure connection
(defmethod execute-command :before (query-string)
  (declare (ignore query-string flatp))
  (ensure-connection *default-database*))

(defgeneric execute-command (sql)
  (:documentation "Executes a SQL command (string or expression)."))

(defmethod execute-command ((sql-string string))
  (db-execute-command *default-database* sql-string))

(defmethod execute-command ((sql-expression sql-command-expression))
  (db-execute-command *default-database* (sql-string sql-expression)))

;; obsolete?
(defparameter *default-query* nil "Default query object.")

(defun open-query (query-expression)
  "Opens a new query and executes query-expression, without closing open queries.
Returns the query object and makes it the default query"
  (when *default-query* ;; should be cerror? 
    (error "There is already a default query."))
  (setf *default-query*
        (db-open-query *default-database* (sql-string query-expression))))

(defun close-query (&optional (query *default-query*))
  "Closes a query."
  (db-close-query query)
  (when (eq query *default-query*)
    (setf *default-query* nil)))

;; I am not quite sure how useful this is. I have tried to cancel a query from a
;; different thread, but this does not seem to work. On the other hand, Command-.
;; (in MCL) does cancel a query (and obviously does not use cancel-query).
;; How does this work? Idea: look into SQLSetConnectOption.
(defun cancel-query (&optional (query-or-database
                                  (or *default-query* *default-database*)))
  "Cancels a query."
  (db-cancel-query query-or-database)
  (setf *default-query* nil))

;; has to be reworked?
(defun fetch-query-results (&key count (database-or-query 
                                            (or *default-query* *default-database*)))
  "Fetches the first <count> results (or all, if count is nil, the default).
The default for database-or-query is (or *default-query* *default-database*)."
  (db-fetch-query-results database-or-query count))

(defun begin-transaction ()
  "Starts a transaction"
  (db-begin-transaction *default-database*))

;; a dynamic variable bound by with-transaction containing the transactions
;; to be committed inside a with-transaction form. It is only used inside 
;; a with-transaction form.
(defvar *active-transactions* () "A list of active transactions")
(defvar *inside-transaction-p* nil "True if we are inside a with-transaction form")

(defparameter *commit-nested-transactions-p* nil
  "If nil, only the outermost with-transaction form does commit on exit.")

;; new transaction control
(defmacro with-transaction ((&optional (database '*default-database*)) &body body)
  "If the forms in body execute without error, all active transactions pertaining
to DATABASE are committed on exit. If an error occurs, the active transactions 
are rolled back.
If *COMMIT-NESTED-TRANSACTIONS-P* is true, transaction are committed/rolled back
on exit from the outermost WITH-TRANSACTION form only."
  (let ((successp (gensym)))
    `(if (and (not *commit-nested-transactions-p*)
              (find ,database *active-transactions*))
       (progn ,@body)
       (let ((,successp nil)
             (*active-transactions*
              (cons ,database *active-transactions*)))
         (db-begin-transaction ,database)
         (unwind-protect
           (multiple-value-prog1
             (progn ,@body)
             (setf ,successp t))
           (if ,successp 
             (db-commit ,database)             
             (db-rollback ,database)))))))

;;; LMH Add hook to allow application to determine the time conversion.
;;; LMH Note possible problem: are fractional seconds allowed in
;;; SQL-92 accessible to the pointer?  This may cause problems because
;;; encode-universal-time need not accept them.
(defparameter *time-conversion-function* (lambda (universal-time &optional fraction)
                                           (declare (ignore fraction))
                                           universal-time)
  "Bound to a function that converts from a Lisp universal time fixnum (and a fractional
as possible second argument) to the desired representation of date/time/timestamp.")

;;; prepared statements

;; works for insert and select statements

(defun prepare-statement (sql-expression &key parameter-columns parameter-table)
  "Prepares a statement for repeated execution and returns a query object. 
Use $ instead of SQL ? as parameter marker. 
Arguments:
 SQL-EXPRESSION : SQL statement to be prepared (Works for now for insert and select).
 PARAMETER-COLUMNS : A list of the parameter columns; only necessary if not all 
are parameterized and if sql-expression is a string. (Otherwise it is possible
to deduce their types and names from the sql-expression.)
 PARAMETER-TABLE : Name of the table; only necessary if sql-expression is a string."
  (db-prepare-statement *default-database* sql-expression
                        :parameter-table parameter-table
                        :parameter-columns parameter-columns))

(defmacro with-prepared-statement ((query sql-expression) &body body)
  "Prepares a (parameterized) statement (insert and select works by now) for repeated
 execution. A new query object is created and bound to the symbol QUERY. 
The function BIND-EXECUTE, to be used inside this macro, binds values to the
parameters and executes the query."
  `(let ((,query (prepare-statement ,sql-expression)))
     (unwind-protect
       (progn ,@body)
       (close-query ,query))))

(defun bind-execute (query &rest parameters)
"Binds values to the parameters of a prepared statement and executes the command.
If the statement execution returns a result set, the result is fetched."
  (apply #'db-execute-parameterized query (sql-expression query) parameters))

(defparameter *null* nil
  "Lisp representation of SQL Null value, default = nil.
May be locally bound to something else if a certain type is necessary.")

(defun select (&rest expressions)
  "Executes a select statement and returns a list of fetched result rows.
Signature: (&rest rows &key database from where group-by having order-by distinct
using set-operation all joins start-with connect-by flatp)
JOINS, USING and SET-OPERATION are not implemented yet" 
  (let ((database (or (cadr (member :database expressions)) *default-database*)))
    (if (null database)
        (error "No database is connected.")
      (let ((query-exp (apply #'make-select-expression database expressions))
            (flatp (cadr (member :flatp expressions))))
        (execute-query query-exp :flatp flatp)))))

(defun select-union (&rest expressions)
  "Executes a union of select statement and returns a list of fetched result rows.
Signature: (&rest select-expressions &key database order-by flatp)"
  (let ((query-exp (apply #'make-union-expression 'union expressions))
        (flatp (cadr (member :flatp expressions))))
    (execute-query (build-sql-string query-exp) :flatp flatp)))

(defun insert-records (&rest rest &key into attributes values av-pairs
                                query-expression &allow-other-keys)
  "Inserts a record into a table."
  (let ((insert-exp
         (make-insert-expression 
          *default-database* into attributes values av-pairs query-expression rest)))
    (execute-command (sql-string insert-exp))))

(defun update-records (table &rest rest &key attributes values av-pairs
                                where &allow-other-keys)
  "Updates a record."
  (let ((update-exp 
         (make-update-expression 
          *default-database* table attributes values av-pairs where rest)))
    (execute-command
     (sql-string update-exp))))

(defun delete-records (&rest rest &key from where &allow-other-keys)
  "Deletes a record."
  (let ((delete-exp (make-delete-expression *default-database* from where rest)))
    (execute-command
     (sql-string delete-exp))))

(defun create-table (table description &rest rest &key &allow-other-keys)
  "Creates a database table.
Arguments:
 TABLE : table expression
 DESCRIPTION : A list of row descriptions and foreign key statements.
A row description is a list of the form (row-name row-type &optional key).
ROW-NAME is a symbol whose symbol-name is the row name, and 
ROW-TYPE is a symbol designating the type of the row or a list of row type and 
precision.
KEY is :NOT-NULL or :PRIMARY-KEY (which implies NOT NULL).
A foreign key statment is a list of the form 
(:foreign-key row-identifier-or-list
 :references foreign-table-identifier foreign-row-identifier-or-list)."
  (execute-command
   (make-create-table-expression *default-database* table description rest)))

(defun drop-table (table)
  "Drops a database table."
  (execute-command
   (make-drop-table-expression *default-database* table)))

(defun create-index (index &rest rest &key on unique-p attributes)
  "Creates an index."
  (execute-command
   (apply #'make-create-index-expression *default-database*
          index on unique-p attributes rest)))

(defun drop-index (index)
  "Drops an index."
  (execute-command
   (make-drop-index-expression *default-database* index)))

(defun create-view (view &rest rest 
                           &key alias as replace-p force-p check-option
                           &allow-other-keys)
  "Creates a database view."
  (execute-command
   (apply #'make-create-view-expression *default-database* view alias replace-p 
          force-p as check-option rest)))

(defun drop-view (view)
  "Drops a database view."
  (execute-command
   (make-drop-view-expression *default-database* view)))

; add password, tablespace etc
(defun create-user (username &rest rest &key password &allow-other-keys)
  "Creates a user (schema)."
  (execute-command
   (apply #'make-create-user-expression *default-database* username password rest)))

(defun execute (&rest rest)
  "Executes a SQL (DDL) command in the generic syntax." 
  (execute-command (apply #'make-generic-execute-expression rest)))

(defun map-query (type function query-exp)
  "Executes a query and maps a function over the result rows. Arguments:
 TYPE : Performs a type coersion. Ignored.
 FUNCTION : A function which as many arguments as there are return columns.
 QUERY-EXP : A select expression."
  (db-map-query *default-database* type function query-exp))

(defun map-bind-query (query type function &rest parameters)
  "Prepared version of MAP-QUERY. Executes a prepared query and maps a
function over the result rows. Arguments:
 QUERY : A prepared select expression.
 TYPE : Performs a type coersion. Ignored.
 FUNCTION : A function which as many arguments as there are return columns.
 PARAMETERS : Parameters to bind for execution."
  (apply #'db-map-bind-query query type function parameters))

;; put the version from sql-streams.lisp back here

(defmacro do-query (((&rest args) query-exp) &rest body)
  "Executes a query; for each result row, the body is executed, while the column 
values are bound to symbols accessible in the body. Arguments:
 ARGS : A list of symbols to bind the result columns to. May contain a &rest 
argument. 
The last argument may be a stream argument of the form (:stream <var>); then, the
symbol <var> is bound to a SQL-STREAM for input which is opened before execution 
of the body and closed afterwards.
 QUERY-EXP : A select expression.
 BODY : body to execute for each row."
  (let* ((last-arg (car (last args)))
         (last-stream-p (and (consp last-arg) (eq :stream (car last-arg))))
         (simple-args (when last-stream-p (butlast args)))
         (stream-arg (when last-stream-p (cadr last-arg)))
         (arglen (let ((rest-pos (position '&rest args)))
                   (cond ((not rest-pos)
                          (length args))
                         ((= rest-pos (- (length args) 2))
                          (if last-stream-p
                            (error "stream and &rest arguments are not allowed together.")
                            nil))
                         (t
                          (error "Malformed argument list ~a." args)))))
         (query (gensym)))
    `(let ((,query (%db-execute *default-database* (sql-string ,query-exp))))
       (unwind-protect
         (progn
           (%initialize-query ,query ,arglen)
           ,(if last-stream-p
              `(let ((,stream-arg (make-sql-stream ,query :column ,(1- arglen)))) 
                 (loop with data and data-found-p
                       do (multiple-value-setq (data data-found-p)
                            (%read-query-data ,query :last))
                       while data-found-p
                       do 
                       (open-sql-stream ,stream-arg)
                       (destructuring-bind ,simple-args data ,@body))
                 (close ,stream-arg))
              `(loop with data
                     while (setf data (%read-query-data ,query nil))
                     do (destructuring-bind ,args data ,@body))))
         ;; dispose of memory and set query inactive or get rid of it
         (db-close-query ,query)))))

(defmacro do-bind-query (((&rest args) query &rest parameters)
                            &rest body)
  "Prepared version of DO-QUERY. Executes a prepared query; for each result row, 
the body is executed, while the column values are bound to symbols accessible
in the body. Arguments:
 ARGS : A list of symbols to bind the result columns to. May contain a &rest 
argument or a stream argument as last argument.
 QUERY : A prepared query object
 PARAMETERS : Parameters to bind for execution.
 BODY : Body to execute for each row."
  (let* ((last-arg (car (last args)))
         (last-stream-p (and (consp last-arg) (eq :stream (car last-arg))))
         (simple-args (when last-stream-p (butlast args)))
         (stream-arg (when last-stream-p (cadr last-arg)))
         (arglen (let ((rest-pos (position '&rest args)))
                   (cond ((not rest-pos)
                          (length args))
                         ((= rest-pos (- (length args) 2))
                          (if last-stream-p
                            (error "stream and &rest arguments are not allowed together.")
                            nil))
                         (t
                          (error "Malformed argument list ~a." args))))))
    `(unwind-protect
       (progn
         (%db-bind-execute ,query ,@parameters)
         ,(if last-stream-p
            `(let ((,stream-arg (make-sql-stream ,query :column ,(1- arglen)))) 
               (loop with data and data-found-p
                     do (multiple-value-setq (data data-found-p)
                          (%read-query-data ,query :last))
                     while data-found-p
                     do 
                     (open-sql-stream ,stream-arg)
                     (destructuring-bind ,simple-args data ,@body))
               (close ,stream-arg))
            `(loop with data
                   while (setf data (%read-query-data ,query nil))
                   do (destructuring-bind ,args data ,@body))))
       ;; dispose of memory and set query inactive or get rid of it
       (%db-reset-query ,query))))

(defun data-sources (&optional (db-type *default-database-type*))
   "Returns a list of (data-source description) - pairs"
   (db-data-sources db-type))

;; DIV 2002-Jun-15  
;; Drop non-generally accepted defaults for catalog (qualifier) or schema (owner) owner.
;; Please leave them for a database-specific db-describe-columns method.
(defmethod describe-columns (&key table-qualifier ;(table-qualifier "")
                                  table-owner ; (table-owner (db-user-id *default-database*))
                                  (table-name "%")
                                  (column-name "%"))
  (db-describe-columns *default-database*
                       (if table-qualifier (string-upcase table-qualifier))
                       (if table-owner (string-upcase table-owner))
                       (string-upcase table-name)
                       (string-upcase column-name)))

(defvar *binary-format* :bit-vector
  "The format binary data is output in. May be one of :bit-vector, :unsigned-byte-vector,
:string, :hex-string")
