;;;-*- Mode: Lisp; Package: SQL -*-

;; SQL module for MCL, LispWorks and ACL
;; Version 0.9
;; Copyright (C) Paul Meurer 1999, 2000. All rights reserved.
;; paul.meurer@hit.uib.no
;;
;; Documentation and the license agreement can be found in file 
;; "sql-odbc-documentation.lisp" 

(in-package :sql)

;; Luckily, both MCL, LW and ACL use closified streams.
;; LW's and ACL's stream implementations are based on the Gray stream proposal.
;; For MCL prior to 4.3, we have to define some classes to adhere to the Gray
;; stream proposal.

#+mcl
(defclass fundamental-stream (stream) ())

#+mcl
(defclass fundamental-input-stream (input-stream fundamental-stream) ())

#+mcl
(defclass fundamental-output-stream (output-stream fundamental-stream) ())

#+mcl
(defclass fundamental-character-stream (fundamental-stream) ())

#+mcl
(defclass fundamental-binary-stream (fundamental-stream) ())

#+mcl
(defgeneric stream-read-char (stream))

;; there is a STREAM-PEEK in MCL
#+mcl
(defgeneric stream-peek-char (stream))

(defclass sql-stream (fundamental-stream)
  ((query :initarg :query :reader stream-query)
   (column :initarg :column :reader sql-stream-column)
   ;; data is read in in chunks from the database. 
   (buffer :initform nil)
   (buffer-size :initform 1024
                :initarg :buffer-size 
                :reader stream-buffer-size)
   ;; read- or write-position in the buffer
   (buffer-position :initform 0))
  (:documentation 
   "Abstract stream class (subclass of FUNDAMENTAL-STREAM) designed primarily for
input and output of long data."))

;; this one has to be subclassed in the odbc (etc.) module
(defclass sql-string-stream (sql-stream fundamental-character-stream) ()
  (:documentation "Subclass of SQL-STREAM for input and output of long string/character data."))

#+maybe
(defclass sql-string-input-stream (sql-string-stream fundamental-input-stream) ())

#+maybe
(defclass sql-string-output-stream (sql-string-stream fundamental-output-stream) ())

#+maybe
(defgeneric db-open-stream (query sql-stream))

#+maybe
(defgeneric db-close-stream (query sql-stream))

(defgeneric make-sql-stream (query &key column stream-class direction
                                      &allow-other-keys))

(defgeneric open-sql-stream (stream))

#+mcl
(defmethod stream-tyi ((stream sql-stream))
  (stream-read-char stream))

#+maybe
(defmethod make-sql-stream ((query query) column
                               &key
                               (stream-class 'sql-string-stream)
                               (direction :output))
  (make-instance stream-class :query query 
                 :column column
                 :direction direction))

#+(and :mcl (not :ccl-4.3))
(defun read-sequence (sequence stream &rest keys &key (start 0) end)
  (declare (dynamic-extent keys))
  (declare (ignore end start))
  (apply 'stream-read-sequence stream sequence keys))

#+(and :mcl (not :ccl-4.3))
(defgeneric stream-read-sequence (stream sequence &key start end))

#+(and :mcl (not :ccl-4.3))
(defgeneric stream-write-sequence (stream sequence &key start end))

#+(and :mcl (not :ccl-4.3))
(defun write-sequence (sequence stream &rest keys &key (start 0) end)
  (declare (dynamic-extent keys))
  (declare (ignore end start))
  (apply 'stream-write-sequence stream sequence keys))

(defmacro bind-execute-with-stream ((query &rest parameters) &body body)
  "Binds values to the parameters of a prepared insert statement and executes the query.
The last parameter value should be a list of the form (:stream <stream-var> [size]),
where <var> is a variable bound to a SQL-STREAM (for output) in the body, and size is 
an optional value denoting the size of the stream. (This is necessary for drivers returning
\"Y\" for info type $SQL_NEED_LONG_DATA_LEN, e.g. the DataDirect Oracle drivers.)"
  (let* (;; list of :stream, stream-var and (optionally) size
         (last-parameter (car (last parameters))) 
         (simple-parameters 
          (progn (assert (and (consp last-parameter)
                              (eq :stream (car last-parameter))))
                 (butlast parameters)))
         (stream-var (cadr last-parameter))
         (new-parameters (append simple-parameters
                                 (list (list 'list :stream (caddr last-parameter))))))
    `(progn
       (db-execute-parameterized ,query (sql-expression ,query) ,@new-parameters)
       (let ((,stream-var (make-sql-stream ,query :column ,(1- (length parameters))
                                           :direction :output)))
         (open-sql-stream ,stream-var) 
         (prog1 ;; unwind-protect? 
           (progn ,@body)
           (close ,stream-var))))))
